module countrateerror_mod
!  VNY 02.11.2004 : a function for rigorous computation
!  of the error bars for the coincidence-loss corrected count rates
!  is added:
!
!  cLossErrBars(frameTime, timeDel, observedCountRate, verbFlag, deadFrac)
!
! The function returns the rms-error estimated with the use of the rigourous algorithm
! for error-bars (S.Rosen, 2004). 
!
! Input parameters:
! ----------------
! frameTime - real(kind=double): frame time in seconds
! timeDel - real: time of the observation, in seconds (it could be a bin-time or the e
! xposure time)
! deadFrac - real(kind=double): dead fraction in units of the fraction of the frame time
! observedCountRate - real : observed count rate (which is affecetd by the coincidence losses)
!
! Output parameters:
!-------------------
! trueCountRate - real: the estimated count rate (corrected for the coincidence losses); 
! this value can be used for checking purporses only
!
!
    use dal
    use cal
    use param
    use errorHandling
    use sasdal
    use omconstants_mod
    use omfunctions_mod
    use splinefit_mod
    use brent_mod
    use omsourceparam_mod
    implicit none

    private initRandomSequence


    public coincLossAsymerr
    public computesimonerrors
   
    public closserrbars1    
    public InitialiseCLossErrBars
    real(single), private, pointer :: clossCountRate(:) => null()

    real(single), private, pointer :: clossErrorBar(:) => null()
    type(OneDSpline), private, save ::  cLossErrBarsSpline
    logical :: initialisedCLossErrBars = .false.
    integer(int16), private, parameter :: maxSize = 100
  
    real(double), private, save :: ec(5) = (/1.0d0, 0.0756730d0, -0.1439860d0, 0.6384970d0, -0.5701840d0/)

    real(double), private :: cLossTrueCountRate, cLossFrameTime1, emp_corr=0.0_double
    real(double), private :: cLossFrameTime, cLossDeadFrac
    real(double), private :: clossCorrCountRate, cLossObservedCountRate
    real(single), private :: cLossExposure
contains

!***********************************************************
!
!
!**********************************************************

subroutine ComputeSimonErrors(nsources,                        & 
                              corrCountRate,                   &
                              corrCountRateErr,                &
                              corrBackgroundCountRate,         & 
                              exposure,                        &
                              deadFrac,                        &
                              frameTime)
    integer(int32) :: nsources
    real(single), pointer, dimension(:) :: corrCountRate, corrCountRateErr
    real(single), pointer, dimension(:) :: corrBackgroundCountRate
    real(single), intent(in) :: exposure
!   real(single), intent(inout) :: exposure

    real(double), intent(in) :: deadFrac, frameTime
    real(single) :: corrBackgroundCountRateErr

    integer(int16) :: i
    real(single) :: rawRate, rawRateErr, totRate, totRateErr, clossCountRate
    real(single) :: rawBackgroundRate, rawBackgroundRateErr
    real(single) :: corrRateERr, corrBackgroundRateErr
    real(single) :: cminRate, cmaxRate


    if (nsources .lt. 1) return

    clossexposure = 1.0
 
    cLossFrameTime = frameTime
    cLossDeadFrac = deadFrac

!exposure = 100.0
!clossFrameTime = 0.01
!clossDeadFrac = 0.0
!corrCountRate = 50.0
    
    if (.not. associated(factorialData%computed)) then

        !*****************************************************
        ! Initialise array for storing log factorials
        !*****************************************************
        call InitialiseFactorial

    end if

    cminRate = 5.0
    cMaxRate = RawFromCorrCountRate(real(0.9999_double / frameTime))

    if (nsources .gt. maxSize) then


        cminRate = max(cminRate, minval(corrCountRate + corrBackgroundCountRate, (corrCountRate+corrBackgroundCountRate) .gt. 0.0))
 


!        cminRate = max(cminRate, minval(corrBackgroundCountRate, corrBackgroundCountRate .gt. 0.0))
 

        cmaxRate = min(cmaxRate, maxval(corrCountRate + corrBackgroundCountRate))
     
        call initialiseClossErrBars(exposure, cminRate, cmaxRate) 
                
        do i = 0, nsources - 1
            totRate = corrCountRate(i) + corrBackgroundCountRate(i)
    
            if (totRate .lt. cminRate .or. totRate .gt. cmaxRate) cycle

 

            !*******************************************************
            ! The error in the total count-rate
            !*******************************************************
             totRateErr = cLossErrBars4( totRate )
   
            !*******************************************************
            ! The error in the background  count-rate
            !*******************************************************
            corrBackgroundCountRateErr = sqrt(corrBackgroundCountRate(i)) * sqrt(exposure) / exposure
            corrCountRateErr(i) = sqrt(totRateErr * totRateErr + &
                 2.0 * corrBackgroundCountRateErr * corrBackgroundCountRateErr)

  
!write(*, *) i, " corrcountrateerr=",   corrCountRateErr(i)

        end do
    else
      
        do i = 0, nsources - 1

            totRate = corrCountRate(i) + corrBackgroundCountRate(i)
    

            if (totRate .lt. cminRate .or. totRate .gt. cmaxRate) cycle



            !**************************************************************
            ! Compute the raw count rate from the corrected count rate
            !**************************************************************

            rawRate = RawFromCorrCountRate( totRate ) 


            !**************************************************************
            ! Compute the error in the corrected count rate 
            !**************************************************************
            totRateErr = cLossErrBars1(clossframeTime, exposure, & 
                              rawRate, clossCountRate, clossdeadFrac)
            

!write(*, *) i, " line 168 totrate=", totrate, " toRateErr=", totRateErr, " exposure=", exposure

            !*******************************************************
            ! The error in the background  count-rate
            !*******************************************************
            corrBackgroundCountRateErr = sqrt(corrBackgroundCountRate(i)) * sqrt(exposure) / exposure
!corrBackgroundcountrateerr = 0.0

            corrCountRateErr(i) = sqrt(totRateErr * totRateErr + &
                 2.0 * corrBackgroundCountRateErr * corrBackgroundCountRateErr)

!write(*, *) i, " line 179 totrate=", totrate, " corrcountrateErr=", corrCountRateErr(i)

       

        end do 
       
    end if



end subroutine ComputeSimonErrors
!********************************************************************
!
!*******************************************************************
subroutine InitialiseCLossErrBars(exposure, cminRate, cmaxrate)


    real(single), intent(in) :: exposure
    real(single), intent(in) :: cminRate, cmaxRate
  
    integer :: i
    real(single) :: dr
    real(single) :: rawCountRate

    if (associated(clossCountRate)) return
    write(errstr,'(a)') "Initialising CLossErrBars"
    call message(AppMsg, Verbose)
  
    dr = (cMaxRate - cminRate) / real(maxSize - 1)

    if (.not. associated(clossErrorBar)) then
        clossCountRate=> allocateSingle(1_int16, maxSize, "clossCountRate")

        clossErrorBar => allocateSingle(1_int16, maxSize, "clossErrorBar")

    end if

    do i = 1, maxSize

        rawCountRate =  cMinRate + real(i - 1) * dr

        ! Compute the error for this count-rate
        
         if (rawCountRate .lt. 1.0) then
             clossCountRate(i) = CorrectedCountRate(dble(rawCountRate), clossframeTime, clossdeadFrac) 
             clossErrorBar(i) = sqrt(clossCountRate(i) * exposure) / exposure
         else
             clossErrorBar(i) = cLossErrBars1(clossframeTime, exposure, & 
                              rawCountRate, clossCountRate(i), clossdeadFrac)

         end if
!         write(*, *) i, " raw rate=", rawCountRate, " error=", clossErrorBar(i),&
!             " corrrate=", clossCountrate(i)

    end do



    ! Construct spline interpolant
    call constructSpline(int(maxSize, int32), cLossCountRate, cLossErrorBar, cLossErrBarsSpline)

    initialisedCLosserrBars = .true.


 end subroutine InitialiseCLossErrBars

!----------------------------------------------------------------------------
! Using Simon's algorithm for CL-correction and estimation of rms
!----------------------------------------------------------------------------
! The function returnes the rms-error for the estimated CL-corrected counts 
! 
!
! Input parameters:
! ----------------
! frameTime - real(kind=double): frame time in seconds
! timeDel - real: time of the observation, in seconds (it could be a bin-time or 
! the exposure time)
! deadFrac - real(kind=double): dead fraction in units of the fraction of the frame time
! observedCountRate - real : observed count rate (which is affecetd by the 
! coincidence losses)
!
! Output parameters:
!-------------------
! trueCountRate - real: the estimated count rate (corrected for the coincidence losses); 
! this value can be used for checking purporses only
!
!
real function cLossErrBars1(frameTime, exposure, & 
     observedCountRate, corrCountRate, deadFrac)
    ! frameTime is in seconds
    real(double), intent(in) :: frameTime
    real(single), intent(in) :: exposure
    real(single), intent(out):: corrCountRate ! counts per second
    real(single), intent(in) :: observedCountRate
    real(double), intent(in) :: deadFrac

    real :: rmsValue
    real(double) :: x, f, mean
    real(double) :: maxCountsPerFrame
    integer :: maxCounts, minCounts
    real :: sim_frameTime, sim_timeBin
    real(double) :: lossCountRate 
    real(double) :: corrRate


    if (frameTime .gt. 0.0) then
        maxCountsPerFrame = 0.9999 / frameTime
    else
        maxCountsPerFrame = 200.0
    endif
      
    maxCounts = 6 * maxCountsPerFrame
    if (maxCounts .gt. 2000) maxCounts=2000

  
    corrRate = CorrectedCountRate(dble(observedCountRate), frameTime, deadFrac)
    corrCountRate = corrRate
    corrRate = corrRate 

    sim_timeBin = exposure

    sim_timeBin = 1.0
    clossExposure = 1.0
    sim_frameTime = frameTime

    !**************************************************************
    ! Check the values of corrRate from corrRate-10% to maxCounts
    !**************************************************************
    minCounts = corrRate - 0.1 * corrRate


    cLossObservedCountRate = observedCountRate / exposure

    x= (maxCounts + minCounts) / 2.0
    f = brent(dble(minCounts), dble(maxCounts), 0.000001_double,  x, meanCountRate)

    call coincLossAsymErr1(sim_frameTime, sim_timeBin, x, &
         lossCountRate, mean)


    call ComputeRmsValue(sim_frameTime, sim_timeBin, x, mean, rmsValue)


    cLossErrBars1 = rmsValue * sqrt(exposure) / exposure

!write(*, *) "Line 351 losscountrate=", losscountrate, " error=", clossErrBars1

end function cLossErrBars1

!*********************************************************
!
!
!********************************************************
real(single) function cLossErrBars4(countRate)
    real(single), intent(in) :: countrate
 

    if (initialisedCLossErrBars) then
        if (countrate .le. clossCountRate(1)) then
            cLossErrBars4 = clossErrorBar(1)
        elseif (countRate .ge. clossCountRate(maxSize)) then
            cLossErrBars4 = clossErrorBar(maxSize)
        else
            clossErrBars4 = interpolateSpline(countRate, cLossErrBarsSpline)
        end if
    else
        if (countRate .ge. 0.0) then
            cLossErrBars4 = sqrt(countRate)
        else
            cLossErrBars4  = 0.0
        end if
  
    end if

end function cLossErrBars4



!*********************************************************
! Function CorrectedCountRate
! Corrects the raw count rate for coincidence losses, etc  
!*********************************************************
real(kind=double) function CorrectedCountRate(rawCountRate, frameTime, deadFrac)
    real(kind=double), intent(in)    :: rawCountRate
    real(kind=double) :: countRateError

    real(kind=double), intent(in) :: frameTime
    real(kind=double), intent(in) :: deadFrac
 
    type(ValueAsymmetricErrorT) :: rawRate
    type(ValueAsymmetricErrorT) :: correctedRate


    countRateError = sqrt(rawCountrate)
    rawRate = ValueAsymmetricErrorT(rawCountRate, dble(countRateError), dble(countRateError))
 
    correctedRate = CAL_omPhotoNatural(rawRate, frameTime, deadFrac, 0.0d0, .false.)

    CorrectedCountRate = CorrectedRate%value
      
end function CorrectedCountRate


!---------------------------------------
! Using Simon's coincidence loss test routine
! 27.05.2004
!---------------------------------------
subroutine coincLossAsymErr(frameTime, timeBin, trueCountRate, &
     lossCountRate, rmsValue)
  real(single), intent(in) :: frameTime
  real(single), intent(in) :: timeBin
  real(single), intent(in) :: trueCountRate ! counts per second
  real(single), intent(out) :: lossCountRate
  
  real(single), intent(out) :: rmsValue
 
  integer :: nFrames


  real(double) :: mu, mu_eff, ln_p
  real(double) :: ln_k_fact, ln_m_fact, ln_k_m_fact, sum_p
  real(double) :: mu_poiss
  real(double) :: rms, mean, r2
  real(double) :: cl_formula, sum_p_l, sum_p_h
  real(double) :: ln_fixed_term, crfrr
  real(double) :: fco

  integer(int32) :: k, i

  real(double) :: ci, t ! incident CR (cts/s) and frame rate (s)
  real(double) :: co_best
  real(double) :: ci_pred, muo
  real(double) :: logoneminusexpminusmu, iminusmean
  real(double) :: f, muBest
  real(double) :: p
  real(double), parameter :: tolerance = 1e-12

! VNY 14.06.2004 from Simon's code:
! insert empirical coincidence loss coefficients, from 
! ~/sasccf/OM_PHOTTONAT_0003.CCF. Coefficients work on the co*t (raw CRFR 
! ratio).


    nFrames = int(timeBin / frameTime)

    ci = trueCountRate
    t = frameTime

    mu_poiss = ci ! incident counts during one second
    mu = ci * t ! incident counts during one frame
    crfrr = ci * t ! INCIDENT CRFR ratio
    cl_formula = (1.d0 - exp(-mu)) / frameTime !t
    k = nFrames ! total number of frames per exposure time (timeBin)

! There is a PROBLEM to use the empirical curve directly here in this 
! test code because the empirical correction uses a polynomial based on 
! the observed, not incident count rate so have to iteratively find what 
! observed CR gives incident count rate provided by user. Quick solution 
! here where we have ci, not co, initially, is to simply compute ci for 
! a range of co*T (=muo =CRFRobserved values) and find value of co which 
! gives Ci (empirical) closest to input value



    !***************************************************
    ! FOR TEST
    !*************************************************

    cLossframeTime1 = t

    clossTrueCountRate = trueCountRate
    mubest = 0.5_double
    f = brent(0.0_double, 0.999999_double, 0.000001_double,  mubest, clecFunction)

    emp_corr = EmpiricalCorrection(muBest)

!write(*, *) "Line 432 f=", f, " mubest=", mubest


    co_best = mubest / t

    ci_pred = -log(1.d0 - mubest) / t
    ci_pred = ci_pred * emp_corr

!write(*, *) "Line 436 co_best=", co_best

    muo = mubest

!**********************************************************************
! END OF TEST
!*********************************************************************




! compute correction factor f(Co) that is thus used to convert this Co to Ci
! Will be used lower down if empirical correction is applied. 

!    muo = co_best * t
    fco = ec(1) + ec(2) * muo + ec(2) * muo * muo + ec(3) * muo**3 + & 
          ec(4) * muo**4 + ec(5) * muo**5
     

! First compute the probability distribution of the incident count rate, for 
! reference
!

  
    sum_p_l = 0.d0
    sum_p_h = 0.d0



!  compute the probability of getting m=i counts if we run k frames with an 
!  incident rate of Ci counts/s and a frame time of t seconds. This 
!  probability is based on kCm * P(0)**(k-m) * P(>=1)**m 
!
!  where
!
!  kCm = k!/(m!)(k-m)!
!
!


    sum_p = 0.d0
    rms = 0.d0
    r2 = 0.d0
    mean = 0.0_double
  
    
    mu_eff = mu / fco   !compute effective incident mu
    cl_formula = co_best
 
    ln_fixed_term = log(1.d0 - dexp(-mu_eff)) !this term if fixed for each k below
  
    ln_k_fact = logFactorial(k)     ! constant for given no of frames


    do i = 0, k  ! m can not exceed k

        ln_m_fact = logFactorial(i)

        ln_k_m_fact = logFactorial(k - i)

        ln_p = ln_k_fact - ln_m_fact - ln_k_m_fact

        ln_p = ln_p - (k - i) * mu_eff + i * ln_fixed_term
        p = dexp(ln_p)

        if(p .lt. 1.d-30) p = 0.d0    ! set to zero if P is very small.
        mean = mean + p * dble(i)
        
        r2 = r2 + p * dble(i * i)
        sum_p = sum_p + p


    end do



!call fatal("stopped")
!--- 15.06.2004
! rounding errors and Stirlings approx give rise to value of sum_p which is
! very slightly > 1.000. Thus, normalise sum(r2) amd mean by sum_p as if to
! make them equivalent to having come from a distribution with sum_p=1.0.
! This appears to remove potential problem of sum(r2)-mean*mean being -ve.

    r2 = r2 / sum_p
    mean = mean / sum_p
    rms = r2 - mean * mean    ! rms (square of)


 !write(*, *) "coincLossAsymErr line 588, sum_p=", sum_p


!
! now compute asymmetric errors - For asymmetric errors need to know
! what mean is first, so must compute above loop first and then repeat to
! get deviations from mean, weighted by P
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


    rms = 0.d0

    logoneminusexpminusmu = log(1.0d0 - dexp(-mu))

    do i = k, 0, -1     ! i can not exceed k

        ln_m_fact = logFactorial(i)

        ln_k_m_fact = logFactorial(k - i)
        ln_p = ln_k_fact - ln_m_fact - ln_k_m_fact
        ln_p = ln_p - dble(k - i) * mu + dble(i) * logoneminusexpminusmu
        p = dexp(ln_p)
        if (p .lt. 1.d-30) p = 0.d0

        iMinusMean = dble(i) - mean
        iMinusMean = iMinusMean * iminusMean

        rms = rms + p * iMinusMean
      
  
  
    end do

  
  
    if (rms .gt. 0.d0) then
        rms = sqrt(rms)

    end if

    rmsValue = rms / timeBin
   
    lossCountRate = mean / timeBin


end subroutine coincLossAsymErr

!******************************************************************
!
! Coincidence-loss emperical correction function
!******************************************************************
real(double) Function clecFunction(muo)
    real(double), intent(in) :: muo

    real(double) :: ci_pred
    real(double) :: muoSquared

    muoSquared = muo * muo

    emp_corr = ec(1) + muo * (ec(1) + ec(2) * muo) + muoSquared * muo * (ec(3)  + &
               ec(4) * muo +  ec(5) * muoSquared)

    ci_pred = (-log(1.d0 - muo) / cLossFrameTime1) * emp_corr

    clecFunction = abs(ci_pred - cLossTrueCountRate)

end function clecFunction


!--------------------------------------
! Initialise the random number generator
!--------------------------------------
subroutine initRandomSequence(initNumber)
  integer, intent(in) :: initNumber
  integer :: seedSize, i
  integer, allocatable :: seedArray(:)
  call random_seed(SIZE=seedSize)
  allocate(seedArray(seedSize))
  do i=1,seedSize
     seedArray(i)=i*initNumber
  enddo
  call random_seed(PUT=seedArray)
  deallocate(seedArray)
end subroutine initRandomSequence




!*************************************************************
! x is the raw count-rate
!************************************************************

real(double) Function CorrToRawRateFunction(x)
    real(double), intent(in) :: x
  
    corrToRawRateFunction = dabs(CorrectedCountRate(x, clossframeTime, 0.0_double) - &
               clossCorrCountRate)
    

end Function CorrToRawRateFunction

!*************************************************************************
!
!
!************************************************************************
real(double) function RawFromCorrCountRate(corrRate)
    real(single), intent(in) :: corrRate
  

    real(double) :: f, xl, xu
 
    clossCorrCountRate = corrRate

    xl = 0.0_double
    xu = 0.99999_double / clossframetime

    RawFromCorrCountRate = (xu - xl) / 2.0

    f = brent(xl, xu, 0.0000001_double,  RawFromCorrCountRate, corrToRawRateFunction)



end function RawFromCorrCountRate


!*************************************************************************
!
!
!************************************************************************
real(double) function MeanCountRate(trueCountRate)
    real(double), intent(in) :: trueCountRate
    real(double) :: x, mean
       
    call coincLossAsymErr1(real(clossFrameTime), cLossExposure, trueCountRate, x, mean)

    meanCountRate = dabs(x - clossObservedCountRate)  


end function MeanCountRate


!---------------------------------------
! Using Simon's coincidence loss test routine
! 27.05.2004
!---------------------------------------
subroutine coincLossAsymErr1(frameTime, timeBin, trueCountRate, &
     lossCountRate, mean)
  real(single), intent(in) :: frameTime
  real(single), intent(in) :: timeBin
  real(double), intent(in) :: trueCountRate ! counts per second
  real(double), intent(out) :: lossCountRate, mean
  
 
  integer :: nFrames


  real(double) :: mu, mu_eff, ln_p
  real(double) :: ln_k_fact, ln_m_fact, ln_k_m_fact, sum_p
  real(double) :: mu_poiss
  real(double) :: rms, r2
  real(double) :: cl_formula, sum_p_l, sum_p_h
  real(double) :: ln_fixed_term, crfrr
  real(double) :: fco

  integer(int32) :: k, i

  real(double) :: ci, t ! incident CR (cts/s) and frame rate (s)
  real(double) :: co_best
  real(double) :: ci_pred, muo
  real(double) :: f, muBest
  real(double) :: p
  real(double), parameter :: tolerance = 1e-12

! VNY 14.06.2004 from Simon's code:
! insert empirical coincidence loss coefficients, from 
! ~/sasccf/OM_PHOTTONAT_0003.CCF. Coefficients work on the co*t (raw CRFR 
! ratio).


    nFrames = int(timeBin / frameTime)

    ci = trueCountRate
    t = frameTime

    mu_poiss = ci ! incident counts during one second
    mu = ci * t ! incident counts during one frame
    crfrr = ci * t ! INCIDENT CRFR ratio
    cl_formula = (1.d0 - exp(-mu)) / frameTime !t
    k = nFrames ! total number of frames per exposure time (timeBin)

! There is a PROBLEM to use the empirical curve directly here in this 
! test code because the empirical correction uses a polynomial based on 
! the observed, not incident count rate so have to iteratively find what 
! observed CR gives incident count rate provided by user. Quick solution 
! here where we have ci, not co, initially, is to simply compute ci for 
! a range of co*T (=muo =CRFRobserved values) and find value of co which 
! gives Ci (empirical) closest to input value



    !***************************************************
    ! FOR TEST
    !*************************************************

    cLossframeTime1 = t

    clossTrueCountRate = trueCountRate

    mubest = 0.5_double
    f = brent(0.0_double, 0.999999_double, 0.000001_double,  mubest, clecFunction)
    emp_corr = EmpiricalCorrection(muBest)


    co_best = mubest / t

    ci_pred = -log(1.d0 - mubest) / t
    ci_pred = ci_pred * emp_corr


     muo = mubest

!**********************************************************************
! END OF TEST
!*********************************************************************


! compute correction factor f(Co) that is thus used to convert this Co to Ci
! Will be used lower down if empirical correction is applied. 

    fco = ec(1) + ec(2) * muo + ec(2) * muo * muo + ec(3) * muo**3 + & 
          ec(4) * muo**4 + ec(5) * muo**5
     

! First compute the probability distribution of the incident count rate, for 
! reference
!

    sum_p_l = 0.d0
    sum_p_h = 0.d0



!  compute the probability of getting m=i counts if we run k frames with an 
!  incident rate of Ci counts/s and a frame time of t seconds. This 
!  probability is based on kCm * P(0)**(k-m) * P(>=1)**m 
!
!  where
!
!  kCm = k!/(m!)(k-m)!
!
!


    sum_p = 0.d0
    rms = 0.d0
    r2 = 0.d0
    mean = 0.0_double
  
    
    mu_eff = mu / fco   !compute effective incident mu
    cl_formula = co_best
 
    ln_fixed_term = log(1.d0 - dexp(-mu_eff)) !this term if fixed for each k below
  
    ln_k_fact = logFactorial(k)     ! constant for given no of frames


    do i = 0, k  ! m can not exceed k

        ln_m_fact = logFactorial(i)

        ln_k_m_fact = logFactorial(k - i)

        ln_p = ln_k_fact - ln_m_fact - ln_k_m_fact

        ln_p = ln_p - (k - i) * mu_eff + i * ln_fixed_term
        p = dexp(ln_p)

        if(p .lt. 1.d-30) p = 0.d0    ! set to zero if P is very small.
        mean = mean + p * dble(i)
        
        r2 = r2 + p * dble(i * i)
        sum_p = sum_p + p

    end do


    mean = mean / sum_p
  
    lossCountRate = mean / timeBin

end subroutine coincLossAsymErr1


!---------------------------------------
! Using Simon's coincidence loss test routine
! 27.05.2004
!---------------------------------------
subroutine ComputeRmsValue(frameTime, timeBin, truecountrate, mean, rmsValue)
    real(single), intent(in) :: frameTime
    real(single), intent(in) :: timeBin
    real(double), intent(in) :: truecountrate, mean
  
    real(single), intent(out) :: rmsValue
 
    integer :: nFrames
    real(double) :: mu, ln_p
    real(double) :: ln_k_fact, ln_m_fact, ln_k_m_fact
    real(double) :: rms
    real(double) :: cl_formula
    real(double) :: crfrr
    real(double) :: mu_poiss

    integer(int32) :: k, i

    real(double) :: ci, t ! incident CR (cts/s) and frame rate (s)
    real(double) :: co_best
    real(double) :: ci_pred, muo
    real(double) :: logoneminusexpminusmu, iminusmean
    real(double) :: f, muBest
    real(double) :: p
    real(double), parameter :: tolerance = 1e-12

    rms = 0.0_double

    nFrames = int(timeBin / frameTime)

    ci = trueCountRate
    t = frameTime

    mu_poiss = ci ! incident counts during one second
    mu = ci * t ! incident counts during one frame

    crfrr = ci * t ! INCIDENT CRFR ratio
    cl_formula = (1.d0 - exp(-mu)) / frameTime !t

    k = nFrames ! total number of frames per exposure time (timeBin)

    cLossframeTime1 = t

    clossTrueCountRate = trueCountRate

    mubest = 0.5_double

    f = brent(0.0_double, 0.999999_double, 0.000001_double,  mubest, clecFunction)

    emp_corr = EmpiricalCorrection(muBest)

    co_best = mubest / t

    ci_pred = -log(1.d0 - mubest) / t
    ci_pred = ci_pred * emp_corr

    muo = mubest

    logoneminusexpminusmu = log(1.0d0 - dexp(-mu))
    ln_k_fact = logFactorial(k)  ! constant for given no of frames



    do i = 0, k     ! i can not exceed k

        ln_m_fact = logFactorial(i)
        ln_k_m_fact = logFactorial(k - i)
        ln_p = ln_k_fact - ln_m_fact - ln_k_m_fact
        ln_p = ln_p - dble(k - i) * mu + dble(i) * logoneminusexpminusmu

        p = dexp(ln_p)
        if (p .lt. 1.d-30) p = 0.d0

        iMinusMean = dble(i) - mean
        iMinusMean = iMinusMean * iminusMean

        rms = rms + p * iMinusMean

    
    end do

  
    if (rms .gt. 0.0_double) then
        rms = sqrt(rms)
        rmsValue = rms / timeBin
  
    else
        rmsValue = 0.0_double

    end if

end subroutine computeRmsValue

!***************************************************************
!
!**************************************************************
real(double) Function EmpiricalCorrection(muo)
    real(double), intent(in) :: muo
    real(double) :: muoSquared, muoCubed

    muoSquared = muo * muo
    muoCubed = muo * muoSquared

    empiricalCorrection = ec(1) + ec(2) * muo + ec(2) * muoSquared + ec(3) * muoCubed + &
                          ec(4) * muoSquared * muoSquared + ec(5) * muoSquared * muoCubed

end Function EmpiricalCorrection

end module countrateerror_mod

